@@ -1431,9 +1431,3 @@ class Neo4j_API(Abclog):
         return int_df.merge(sifts_df, how='left').merge(i3d.unp_chain_pairs_p(pdbs), how='left').merge(i3d.all_unp_pairs_12, how='left')
 
 
-'''
-TODO: 
-    1. Append Partner SIFTS Info (In Half Way)
-    2. Score SIFTS
-    3. SELECT SIFTS
-'''
